"""
Todo:
- check for oververtex: checks whether a mesh with modifiers applied and uv/vcol vertex splitting is over 0x8000/0xffff vertices
- rename the tool panel to something better

#"""



import bpy, mathutils, bmesh, os, time
#from bpy.props import *
#from bl_ui import properties_data_modifier,space_info
#from math import sqrt


def register():
    bpy.utils.register_class(SortSelectedByLayerAndPO)
    bpy.utils.register_class(SelectionAsPO)
    bpy.utils.register_class(SelectionAsNPO)
    bpy.utils.register_class(SelectAllPO)
    bpy.utils.register_class(SelectAllNPO)
    bpy.utils.register_class(SelectOnlyPO)
    bpy.utils.register_class(SelectOnlyNPO)
    bpy.utils.register_class(SelectAllMeshes)
    bpy.utils.register_class(SelectAllCameras)
    bpy.utils.register_class(SelectAllLights)
    bpy.utils.register_class(SelectedInActivePage)
    bpy.utils.register_class(MakeSelectedPhysical)
    bpy.utils.register_class(MakeSelectedTerrain)
    bpy.utils.register_class(SampleKeyframes)
    bpy.utils.register_class(ScaleAll)
    bpy.utils.register_class(KormanToolsPanel)

def unregister():
    bpy.utils.unregister_class(KormanToolsPanel)
    bpy.utils.unregister_class(SortSelectedByLayerAndPO)
    bpy.utils.unregister_class(SelectionAsPO)
    bpy.utils.unregister_class(SelectionAsNPO)
    bpy.utils.unregister_class(SelectOnlyPO)
    bpy.utils.unregister_class(SelectOnlyNPO)
    bpy.utils.unregister_class(SelectAllMeshes)
    bpy.utils.unregister_class(SelectAllCameras)
    bpy.utils.unregister_class(SelectAllLights)
    bpy.utils.unregister_class(SelectedInActivePage)
    bpy.utils.unregister_class(MakeSelectedPhysical)
    bpy.utils.unregister_class(MakeSelectedTerrain)
    bpy.utils.unregister_class(ScaleAll)
    bpy.utils.unregister_class(SampleKeyframes)



pass
'''
class SortSelectedByLayerAndPO(bpy.types.Operator):
    """Puts PO on top row of layers, and non PO on bottom row"""
    bl_idname = "korman_operators.sort_selected_by_layer_and_po"
    bl_label = "Sort PO from non PO"

    def execute(self, context):
        for obj in context.scene.objects:
            if obj.select:
                lyrs = [i for i in range(20) if obj.layers[i]]
                if len(lyrs) > 1: continue
                if lyrs[0] >= 10 and obj.plasma_object.enabled:
                    obj.layers[lyrs[0] % 10] = True
                    obj.layers[lyrs[0]] = False
                elif lyrs[0] < 10 and not obj.plasma_object.enabled:
                    obj.layers[lyrs[0]+10] = True
                    obj.layers[lyrs[0]] = False
        return{'FINISHED'}'''


# newer version (after looking at Korman code to see which behavior works best with the plugin's paradigm)
class SortSelectedByLayerAndPO(bpy.types.Operator):
    """Puts PO on layer corresponding to page id used. Non PO are either not touched or moved to the last layer. PO without pages are put on last-but-one layer.
    Assuming the Age has less than 18 pages."""
    bl_idname = "korman_operators.sort_selected_by_layer_and_po"
    bl_label = "Sort PO from non PO"

    def execute(self, context):
        world = context.scene.world
        pages = {page.name: page.seq_suffix for page in world.plasma_age.pages}
        for obj in context.scene.objects:
            if obj.type == "LAMP":
                # Lamp setup will be screwed either way. So let's ignore it.
                print("Not moving lamp {}, do it yourself !".format(obj.name))
                continue
            if obj.plasma_object.enabled:
                pageName = obj.plasma_object.page
                if pageName:
                    if pageName in pages:
                        # object is PO and has valid page name. Moving it.
                        normalLyr = pages[pageName] - 1 # since 0 is default page, which is hidden
                        obj.layers = [(False, True)[i==normalLyr] for i in range(20)]
                    else:
                        # object is PO but page name is invalid. Let's just put it on default page.
                        obj.plasma_object.page = ""
                if not pageName:
                    # object has no page, or page was invalid. Putting it on default page (which we'll put on layer 18.
                    obj.layers = [(False, True)[i==18] for i in range(20)]
            else:
                # object isn't a PRP object (and not a lamp). Whatever.
                obj.layers = [(False, True)[i==19] for i in range(20)]
        return{'FINISHED'}


class SelectionAsPO(bpy.types.Operator):
    """Makes selection Plasma Objects"""
    bl_idname = "korman_operators.selection_as_po"
    bl_label = "Selection to PO"

    def execute(self, context):
        for obj in context.scene.objects:
            if obj.select:
                obj.plasma_object.enabled = True
        return{'FINISHED'}


class SelectionAsNPO(bpy.types.Operator):
    """Makes selection non-Plasma objects"""
    bl_idname = "korman_operators.selection_as_npo"
    bl_label = "Selection to non PO"

    def execute(self, context):
        for obj in context.scene.objects:
            if obj.select:
                obj.plasma_object.enabled = False
        return{'FINISHED'}


class SelectAllPO(bpy.types.Operator):
    """Add all Plasma Objects on current layers to selection"""
    bl_idname = "korman_operators.select_all_po"
    bl_label = "Select PO"

    def execute(self, context):
        for obj in context.scene.objects:
            if obj.plasma_object.enabled:
                for i in range(20):
                    if obj.layers[i] and context.scene.layers[i]:
                        obj.select = True
        return{'FINISHED'}


class SelectAllNPO(bpy.types.Operator):
    """Add all non Plasma Objects on current layers to selection"""
    bl_idname = "korman_operators.select_all_npo"
    bl_label = "Select non PO"

    def execute(self, context):
        for obj in context.scene.objects:
            if not obj.plasma_object.enabled:
                for i in range(20):
                    if obj.layers[i] and context.scene.layers[i]:
                        obj.select = True
        return{'FINISHED'}


class SelectOnlyPO(bpy.types.Operator):
    """Select only Plasma Object on the current layers"""
    bl_idname = "korman_operators.select_only_po"
    bl_label = "Select only PO"

    def execute(self, context):
        for obj in context.scene.objects:
            for i in range(20):
                if obj.layers[i] and context.scene.layers[i]:
                    obj.select = obj.plasma_object.enabled
        return{'FINISHED'}


class SelectOnlyNPO(bpy.types.Operator):
    """Select only non Plasma Object on the current layers"""
    bl_idname = "korman_operators.select_only_npo"
    bl_label = "Select only non PO"

    def execute(self, context):
        for obj in context.scene.objects:
            for i in range(20):
                if obj.layers[i] and context.scene.layers[i]:
                    obj.select = not obj.plasma_object.enabled
        return{'FINISHED'}


class SelectAllMeshes(bpy.types.Operator):
    """Add meshes to selection"""
    bl_idname = "korman_operators.select_all_meshes"
    bl_label = "Select meshes"

    def execute(self, context):
        for obj in context.scene.objects:
            if obj.type == "MESH":
                obj.select = True
        return{'FINISHED'}


class SelectAllCameras(bpy.types.Operator):
    """Add cameras to selection"""
    bl_idname = "korman_operators.select_all_cameras"
    bl_label = "Select cameras"

    def execute(self, context):
        for obj in context.scene.objects:
            if obj.type == "CAMERA":
                obj.select = True
        return{'FINISHED'}


class SelectAllLights(bpy.types.Operator):
    """Add lights to selection"""
    bl_idname = "korman_operators.select_all_lights"
    bl_label = "Select lights"

    def execute(self, context):
        for obj in context.scene.objects:
            if obj.type == "LAMP":
                obj.select = True
        return{'FINISHED'}


class SelectedInActivePage(bpy.types.Operator):
    """Puts all selected objects in the same page as the active object's page"""
    bl_idname = "korman_operators.selected_in_active_page"
    bl_label = "Page to active"

    def execute(self, context):
        for obj in context.scene.objects:
            if obj.select:
                obj.plasma_object.page = context.object.plasma_object.page
        return{'FINISHED'}


class MakeSelectedPhysical(bpy.types.Operator):
    """Add collision to all selected objects"""
    bl_idname = "korman_operators.make_selected_physical"
    bl_label = "Make collider"

    def execute(self, context):
        for obj in context.scene.objects:
            if obj.select:
                context.scene.objects.active = obj
                bpy.ops.object.plasma_modifier_add(types="collision")
                obj.plasma_modifiers.collision.bounds = "trimesh"
                
        return{'FINISHED'}


class MakeSelectedTerrain(bpy.types.Operator):
    """Add collisions as terrain to all selected objects"""
    bl_idname = "korman_operators.make_selected_terrain"
    bl_label = "Make terrain"

    def execute(self, context):
        for obj in context.scene.objects:
            if obj.select:
                context.scene.objects.active = obj
                bpy.ops.object.plasma_modifier_add(types="collision")
                obj.plasma_modifiers.collision.terrain = True
                obj.plasma_modifiers.collision.bounds = "trimesh"
        return{'FINISHED'}


class ScaleAll(bpy.types.Operator):
    """Scale the whole scene by 43.16 (factor can't be modified yet)"""
    bl_idname = "korman_operators.scale_all"
    bl_label = "Scale by 43.16"

    def execute(self, context):
        factor = 43.16
        #print("WARNING - to be implemented in a better way !")
        # offsets the object's location
        # scales each vertex relative to origin
        # scales each animation curve related to loc or delta loc
        # is that all ? hope so.
        processed = []
        processedActions = []
        processedLights = []
        mist = context.scene.world.mist_settings
        mist.depth *= factor
        mist.start *= factor
        mist.height *= factor
        for obj in context.scene.objects:
            if not obj.parent:
                obj.location.x *= factor
                obj.location.y *= factor
                obj.location.z *= factor
            else:
                # a bit special, since we have two matrices to update
                obj.matrix_basis[0][3] *= factor
                obj.matrix_parent_inverse[0][3] *= factor
                obj.matrix_basis[1][3] *= factor
                obj.matrix_parent_inverse[1][3] *= factor
                obj.matrix_basis[2][3] *= factor
                obj.matrix_parent_inverse[2][3] *= factor
            if obj.type == "MESH":
                if not (obj.data.name in processed):
                    for vert in obj.data.vertices:
                        vert.co.x *= factor
                        vert.co.y *= factor
                        vert.co.z *= factor
                    processed.append(obj.data.name)
            if obj.type == "LAMP":
                if not (obj.data.name in processedLights):
                    obj.data.distance *= factor
                    processedLights.append(obj.data.name)
            if obj.animation_data and obj.animation_data.action:
                if obj.animation_data.action.name in processedActions:
                    continue
                for curve in obj.animation_data.action.fcurves:
                    if curve.data_path == "location" or curve.data_path == "delta_location":
                        for keyfrm in curve.keyframe_points:
                            keyfrm.co[1] *= factor
                            # don't forget handles need scaling as well
                            keyfrm.handle_left[1] *= factor
                            keyfrm.handle_right[1] *= factor
                        #curve.update() # ensures handles don't go funny
                processedActions.append(obj.animation_data.action.name)
        return{'FINISHED'}


class SampleKeyframes(bpy.types.Operator):
    """Samples keyframes for the selected objects. Because Blender's approach is wrong."""
    bl_idname = "korman_operators.sample_keyframes"
    bl_label = "Sample keyframes"

    def execute(self, context):
        for obj in context.scene.objects:
            if not obj.select:
                continue
            
            curves = {}
            
            if obj.animation_data and obj.animation_data.action:
                for curve in obj.animation_data.action.fcurves:
                    if not (curve.data_path in curves):
                        curves[curve.data_path] = {}
                    curves[curve.data_path][curve.array_index] = {}
                    bc = curves[curve.data_path][curve.array_index]
                    for i in range(int(curve.keyframe_points[-1].co[0])):
                        bc[i] = curve.evaluate(i)
                
                for curve in obj.animation_data.action.fcurves:
                    # remove
                    while len(curve.keyframe_points) != 0:
                        curve.keyframe_points.remove(curve.keyframe_points[-1])
                # append back
                for curve_data_path in curves:
                    curve = curves[curve_data_path]
                    for curve_array_index in curve:
                        points = curve[curve_array_index]
                        for blc in obj.animation_data.action.fcurves:
                            if blc.data_path == curve_data_path and blc.array_index == curve_array_index:
                                for time in points:
                                    blc.keyframe_points.add()
                                    blc.keyframe_points[-1].interpolation = "LINEAR"
                                    blc.keyframe_points[-1].co[0] = time
                                    blc.keyframe_points[-1].co[1] = points[time]
                                break
        
        return{'FINISHED'}


class CleverLoD(bpy.types.Operator):
    """Creates multiple objects with different visibility distances from 1 big object, automatically splitting loose parts.
    Obviously it doesn't do that yet, it will require to be integrated into Korman itself. Just some template code."""
    bl_idname = "korman_operators.clever_lod"
    bl_label = "Clever LoD"
    
    def execute(self, context):
        # read options
        if obj.plasma_options.autolod:
            bpy.ops.mesh.separate(type='LOOSE')
            for obj in createdobjs:
                # apply a decimate modifier to each smaller object
                # then set mindist and maxdist
                
                # wait. That's stupid. There is a simpler way...
                # Since most of the time we're dealing with duplicate objects created with array
                # modifiers, or even two objects sharing the same mesh, we can simply create
                # a plClusterGroup of these objects. These use depth-based alpha by default, which is
                # what we want.
                
                # However, this means taking modifiers into account, thus impossible to implement using
                # my lightbaking script along with Korman.
                
                # Hmmmm... Just keep the idea around for now.
                pass
        return {'FINISHED'}





class KormanToolsPanel(bpy.types.Panel):
    """Panel integrated into Blender's UI"""
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'TOOLS'
    bl_label = "Korman tools"

    def draw(self,context):
        layout = self.layout
        view = context.object
        layout.operator("korman_operators.sort_selected_by_layer_and_po", text="Sort PO from non PO")
        layout.operator("korman_operators.selection_as_po", text="Selected to PO")
        layout.operator("korman_operators.selection_as_npo", text="Selected to non PO")
        layout.operator("korman_operators.select_all_po", text="Select PO")
        layout.operator("korman_operators.select_all_npo", text="Select non PO")
        layout.operator("korman_operators.select_only_po", text="Select only PO")
        layout.operator("korman_operators.select_only_npo", text="Select only non PO")
        layout.operator("korman_operators.select_all_meshes", text="Select meshes")
        layout.operator("korman_operators.select_all_cameras", text="Select cameras")
        layout.operator("korman_operators.select_all_lights", text="Select lights")
        layout.operator("korman_operators.selected_in_active_page", text="Page to active")
        layout.operator("korman_operators.make_selected_physical", text="Make collider")
        layout.operator("korman_operators.make_selected_terrain", text="Make terrain")
        layout.operator("korman_operators.sample_keyframes", text="Sample keyframes")
        # layout.operator("korman_operators.scale_all", text="Scale by 43.16")
